/*
 * @(#) Current.java
 *
 * JOTM: Java Open Transaction Manager 
 *
 *
 * This module was originally developed by 
 *
 *  - Bull S.A. as part of the JOnAS application server code released in 
 *    July 1999 (www.bull.com)
 * --------------------------------------------------------------------------
 *  The original code and portions created by Bull SA are 
 *  Copyright (c) 1999 BULL SA  
 *  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------- 
 *
 * Contributor(s):
 * 01/11/06 Christophe Ney cney@batisseurs.com 
 *          Added ResourceManagerListener mechanism to remove ThreadData
 *          dependency.
 * 01/12/03 Dean Jennings - synchronizedMap for txXids
 *
 * --------------------------------------------------------------------------
 * $Id: Current.java,v 1.61 2006-09-06 20:44:58 tonyortiz Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.EmptyStackException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.resource.spi.XATerminator;
import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.transaction.InvalidTransactionException;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.UserTransaction;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;

import org.objectweb.transaction.jta.ResourceManagerEvent;
import org.objectweb.transaction.jta.TransactionManager;
import org.objectweb.howl.log.xa.XACommittingTx;

/**
 * <code>Current</code> is the common implementation for both
 * <code>UserTransaction</code> and <code>TransactionManager</code>.
 * <ul> 
 * <li><code>UserTransaction</code> is used by clients that want to demarcate
 * transactions themselves. It is referenceable through JNDI.</li>
 * <li><code>TransactionManager</code> is used by an application server.</li>
 * </ul>
 * 
 * <p>This object is unique in a VM, i. e. each application server has
 * <em>ONE</em> <code>Current</code> object and each client program should
 * normally issue only <em>ONE</em> lookup on JNDI.</p>
 * 
 * <p><code>Current</code> also implements <code>Referenceable</code> and
 * <code>Serializable</code> because of JNDI.</p>
 */

public class Current implements UserTransaction, TransactionManager, Referenceable, Serializable {

    // transaction associated to the thread
    // Thread specific data
    private transient static ThreadLocal threadTx = new ThreadLocal();
    private transient static ThreadLocal threadTimeout = new ThreadLocal();

    // ------------------------------------------------------------------
    // Private data
    // ------------------------------------------------------------------

    // Static hashtable: Xid ---> transaction
    private transient static Map txXids = Collections.synchronizedMap(new HashMap());

    // Must be init at null, for clients that do not get UserTransaction
    private transient static Current unique = null;
    
    // private info	
    private transient static TimerManager timermgr = null; // local
    private transient static TransactionFactory tm = null; // local or remote
    private transient static TransactionRecovery tr = null; // transaction recovery object

    private static final String JOTM_VERSION = "JOTM 2.0.13";
    private static final int DEFAULT_TIMEOUT = 60;
    private int defaultTimeout = DEFAULT_TIMEOUT;
    // private int transactionTimeout = DEFAULT_TIMEOUT;
    private static final boolean DEFAULT_RECOVERY = false;
    private static boolean transactionRecovery = DEFAULT_RECOVERY;
    private static boolean appServer = true;
    
    // management counter
    private transient int nb_bg_tx = 0;
    private transient int nb_rb_tx = 0;
    private transient int nb_cm_tx = 0;
    private transient int nb_to = 0;

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * Default constructor.
     * A client does not need the TMFactory.
     */

    // private Current(){  for unique instance
    public Current(){
        TraceTm.jta.info(JOTM_VERSION);
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("no args constructor");
        }

        unique = this;

        timermgr = TimerManager.getInstance();
        
        if (tr == null) {
            try {
                tr = new TransactionRecoveryImpl ();
            } catch (Exception e) {
                setDefaultRecovery(false);
                TraceTm.recovery.error("Cannot open Howl Log");
                TraceTm.recovery.error("JOTM Recovery is being disabled");
            }
        }
    }

    /**
     * Constructor for an application server. 
     * 
     * The TM factory is passed as an argument. Note that the TM factory can be
     * either local or remote.
     * 
     * @param tmfact TM Factory to use
     */

    public Current(TransactionFactory tmfact){
        TraceTm.jta.info(JOTM_VERSION);
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("TransactionFactory="+ tmfact);
        }
        
        unique = this;
        
        tm = tmfact;
        
        timermgr = TimerManager.getInstance();
        
        if (tr == null) {
           try {
               tr = new TransactionRecoveryImpl ();
           } catch (Exception e) {
               if (TraceTm.jta.isDebugEnabled()) {
                   TraceTm.jta.debug("Exception from TransactionRecoveryImpl= "+ e);
               }
               setDefaultRecovery(false);
               TraceTm.recovery.error("Cannot open Howl Log");
               TraceTm.recovery.error("JOTM Recovery is being disabled");
           }
       }
    }
    
    /**
     * Returns the unique instance of the class.
     *
     * @return The <code>Current</code> object created 
     */
    
    // public static synchronized Current getInstance() {
    //     if (unique == null) {
    //         unique = new Current();
    //     }
        
    //     if (TraceTm.jta.isDebugEnabled()) {
    //         TraceTm.jta.debug("Current.getInstance unique= " + unique);
    //     }

    //     return unique;
    // }
    
    /**
     * The TM factory is passed as an argument. Note that the TM factory can be
     * either local or remote.
     * 
     * @param tmfact TM Factory to use
     */
    
    public static void setTMFactory (TransactionFactory tmfact) {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("TransactionFactory="+ tmfact);
        }
        
        tm = tmfact;   
    }
    
    /**
     * Initiate transaction recovery if enabled.
     */
    
    public static void transactionRecovery () {
        if (tr == null) {
            try {
                tr = new TransactionRecoveryImpl ();
            } catch (Exception e) {
                setDefaultRecovery(false);
                TraceTm.recovery.error("Cannot open Howl Log");
                TraceTm.recovery.error("JOTM Recovery is being disabled");
            }
        }
    }
    
    /**
     * Gets the <code>TransactionManager</code> instance.
     * 
     * @return TransactionManager
     */

    public static TransactionManager getTransactionManager() {
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("getTransactionManager");
        }
        
        return (TransactionManager) unique;
    }

    // ------------------------------------------------------------------
    // UserTransaction implementation
    // ------------------------------------------------------------------

    /**
     * Creates a new transaction and associate it with the current thread.
     *
     * @exception NotSupportedException Thrown if the thread is already
     *    associated with a transaction. (nested transaction are not
     *    supported)
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */

    public void begin() throws NotSupportedException, SystemException {

        int transactionTimeout = 0;
        List templist = null;
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("begin transaction");
        }

        // checks that no transaction is already associated with this thread.
        TransactionImpl tx = (TransactionImpl) threadTx.get();

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.get= " + threadTx.toString());
        }

        if (tx != null) {
            if (txXids.containsValue(tx)) {
                throw new NotSupportedException("Nested transactions not supported");
            } else {
                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("Resetting current tx = " + tx + " since it is already completed.");
                }
            }
        }

        Integer mytimeobj = (Integer) threadTimeout.get();
        
        if (mytimeobj == null) { // no setTransactionTimeout, use defaultTimeout
            transactionTimeout = defaultTimeout;   
        } else {
            if (mytimeobj.intValue() == 0) {  // never set, use defaultTimeout
                transactionTimeout = defaultTimeout;   
            } else {
                transactionTimeout = mytimeobj.intValue();
                threadTimeout.set(new Integer(0));  // reset to "not set", integer 0
            }
        }
        
        // builds a new Xid
        // - should pass servername + ip addr. (LATER)
        XidImpl otid = new XidImpl();

        // creates a new TransactionImpl object
        // - May raise SystemException
        tx = new TransactionImpl(otid, transactionTimeout);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("tx=" + tx);
        }

        /// clear suspended transaction. 
        try {
            tx.doAttach(XAResource.TMJOIN);
        } catch (RollbackException e) {
            // never.
            TraceTm.jotm.error("doAttach: RollbackException");
            throw new SystemException("RollbackException in occured in begin() " + e.getMessage());
        }

        // associates transaction with current thread
        threadTx.set(tx);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.set= " + threadTx.toString());
        }

        // associates this Tx with the Xid
        putTxXid(otid, tx);

        // sets a timer for the transaction
        if (timermgr != null) {
            tx.setTimer(timermgr.addTimer(tx, transactionTimeout, null, false));
        }
        
        // sets the time stamp for the transaction
        Date myDate = new Date();
        tx.setTxDate(myDate);

        // enlist all connections using the event callback this should be done in UserTransaction
        // was done in doAttach
        Stack curStack = (Stack) eventListStack.get();

        if (curStack != null) { // if there is opened connections

            try {
                List list = (List) curStack.peek();

                if (list != null) {
                    templist = new Vector(list); // avoid using an iterator on global list
                    for (Iterator it = templist.iterator(); it.hasNext();) {
                        ((ResourceManagerEvent) it.next()).enlistConnection(tx);
                    }
                } else {

                    if (TraceTm.jta.isDebugEnabled()) {
                        TraceTm.jta.debug("Current.begin called with null list");
                    }
                }
            } catch (EmptyStackException e) {
                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("Current.begin called with empty stack");
                }
            }
        }
    }

    // ------------------------------------------------------------------
    // Inflow Transaction implementation
    // ------------------------------------------------------------------

    /**
     * Creates a new inflow transaction and associates it with the current thread.
     *
     * @param passxid <code>Xid</code> of the inflow transaction.
     *
     * @exception NotSupportedException Thrown if the thread is already
     *    associated with a transaction. (nested transaction are not
     *    supported)
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */

    public void begin(javax.transaction.xa.Xid passxid) throws NotSupportedException, SystemException {
        
        Integer mytimeobj = (Integer) threadTimeout.get();
        
        if (mytimeobj == null) {  // setTranactionTimeout not called, use defaultTimeout
            begin(passxid, (long) defaultTimeout);   
        } else {
            if (mytimeobj.intValue() == 0) {  // transactionTimeout not set, use defaultTimeout
                begin(passxid, (long) defaultTimeout);   
            } else {
                begin(passxid, (long) (mytimeobj.intValue()));
                threadTimeout.set(new Integer(0));  // resets to "not set", integer 0
            }
        }
        // begin(passxid, (long) transactionTimeout);
    }

    /**
     * Creates a new inflow transaction and associates it with the current thread.
     *
     * @param passxid <code>Xid</code> of the inflow transaction.
     *
     * @param timeout value of the timeout (in seconds). If the value is less than 
     * or equal to zero, the value will be set to the default value.
     *
     * @exception NotSupportedException Thrown if the thread is already
     *    associated with a transaction. (nested transaction are not
     *    supported)
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */

    public void begin(javax.transaction.xa.Xid passxid, long timeout) throws NotSupportedException, SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("begin inflow transaction, timeout = " + timeout);
        }

        if (timeout <= 0) {
            timeout = defaultTimeout;
        }
        
        // checks that no transaction is already associated with this thread.
        TransactionImpl tx = (TransactionImpl) threadTx.get();

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.get= " + threadTx.toString());
        }

        if (tx != null) {
            if (txXids.containsValue(tx)) {
                throw new NotSupportedException("Nested transactions not supported");
            } else {
                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("Resetting current tx = " + tx + " since it is already completed.");
                }
            }
        }

        // stores the passed xid components
        XidImpl pxid = new XidImpl(passxid);

        // creates a new TransactionImpl object
        // - May raise SystemException
        tx = new TransactionImpl(pxid, (int) timeout);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("tx=" + tx);
        }

        // associates transaction with current thread
        threadTx.set(tx);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.set= " + threadTx.toString());
        }

        // associates this Tx with the Xid
        putTxXid(pxid, tx);

        // sets a timer for the transaction
        if (timermgr != null) {
            tx.setTimer(timermgr.addTimer(tx, (int) timeout, null, false));
        }

        // sets the time stamp for the transaction
        Date myDate = new Date();
        tx.setTxDate(myDate);

    }

    /**
     * Gets the inflow transaction object that represents the transaction context of
     * the calling thread.
     *
     * @return the XATerminator object representing the inflow transaction 
     * associated with the calling thread. If the calling thread is
     * not associated with an inflow transaction, a null object reference
     * is returned.
     *
     * @exception XAException Thrown if the transaction manager 
     * encounters an unexpected error condition
     */

    public XATerminator getXATerminator() throws XAException {

        XATerminator xaterm = null;

        try {
            xaterm = new XATerminatorImpl();
        } catch (XAException e) {
            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("Cannot create XATerminatorImpl"+ e);
            }
        }

        return xaterm;
    }

    // ------------------------------------------------------------------
    // End of Inflow Transaction implementation
    // ------------------------------------------------------------------


    /**
      * Commits the transaction associated with the current thread. When this
      * method completes, the thread becomes associated with no transaction.
      *
      * @exception RollbackException Thrown to indicate that    the transaction
      * has been rolled back rather than committed.
      *
      * @exception HeuristicMixedException Thrown to indicate that a heuristic
      * decision was made and that some relevant updates have been committed
      * while others have been rolled back.
      *
      * @exception HeuristicRollbackException Thrown to indicate that a
      * heuristic decision was made and that some relevant updates have been
      * rolled back.
      *
      * @exception SecurityException Thrown to indicate that the thread is not
      * allowed to commit the transaction.
      *
      * @exception IllegalStateException Thrown if the current thread is    not
      * associated with a transaction.
      *
      * @exception SystemException Thrown if the transaction manager encounters
      * an unexpected error condition
      */

    public void commit()
        throws
            RollbackException,
            HeuristicMixedException,
            HeuristicRollbackException,
            SecurityException,
            IllegalStateException,
            SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("commit transaction ");
        }

        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();

        if (tx == null) {
            throw new IllegalStateException("Cannot get Transaction for commit");
        }

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("tx=" + tx);
        }
        
        // Commit Transaction. Exceptions may be raised!
        try {
            tx.commit();
        } finally {
            // Dissociates the transaction from current thread
            // Has not been done in doDetach because we need to
            // be in the transactional context for beforeCompletion
            threadTx.set(null);
            
            // Reset the timeout to the default timeout
            // transactionTimeout = defaultTimeout;
            
            // if (TraceTm.jta.isDebugEnabled()) {
            //     TraceTm.jta.debug("threadTx.set= null");
            //     TraceTm.jta.debug("reset timeout= " + defaultTimeout);
            // }
        }
    }

    /**
     * Rolls back the transaction associated with the current thread. When this
     * method completes, the thread becomes associated with no transaction.
     *
     * @exception SecurityException Thrown to indicate that the thread is    not
     * allowed to roll back the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is    not
     * associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */
    public void rollback() throws IllegalStateException, SecurityException, SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.rollback()");
        }

        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();

        if (tx == null) {
            throw new IllegalStateException("Cannot get Transaction for rollback");
        }
        
        threadTx.set(null);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.set= null");
        }

        // Roll back the transaction. Exceptions may be raised!
        tx.rollback();
        
        // if (TraceTm.jta.isDebugEnabled()) {
        //     TraceTm.jta.debug("reset timeout= " + defaultTimeout);
        // }
        
        // transactionTimeout = defaultTimeout;
        
    }

    /**
     * Modify the transaction associated with the current thread such that the
     * only possible outcome of the transaction is to roll back the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is    not
     * associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */

    public void setRollbackOnly() throws IllegalStateException, SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.setRollbackOnly()");
        }
        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();

        if (tx == null) {
            throw new IllegalStateException("Cannot get Transaction for setRollbackOnly");
        }

        // Set transaction rollback only. Exceptions may be raised!
        tx.setRollbackOnly();
    }

    /**
      * Returns the status of the transaction associated with the current
      * thread.
      * 
      * @return transaction status. If no transaction is associated with the
      * current thread, this method returns the Status.NoTransaction value.
      *
      * @exception SystemException Thrown if the transaction manager
      * encounters an unexpected error condition
      */

    public int getStatus() throws SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.getStatus()");
        }

        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();

        if (tx == null) {
            return Status.STATUS_NO_TRANSACTION;
        }

        // Get TX status. Exceptions may be raised!
        return tx.getStatus();
    }

    /**
      * Modifies the value of the timeout value that is associated with the
      * transactions started by the current thread with the begin method.
      *
      *  If    an application has not called this method, the transaction
      * service uses some default value for the transaction timeout.
      *
      * @param timeout value of the timeout (in seconds). If the value is zero,
      * the transaction service restores the default value.
      *
      * @exception SystemException Thrown if the transaction manager
      * encounters an unexpected error condition
      */

    public void setTransactionTimeout(int timeout) throws SystemException {
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("timeout= "+ timeout);
        }
        
        // checks that no transaction is already associated with this thread.
        // If one is, then we have a running transaction (ut.begin) and we must
        // wait until transaction completes (ut.commit or ut.rollback)
        
        TransactionImpl tx = (TransactionImpl) threadTx.get();

        if (tx != null) {
            if (txXids.containsValue(tx)) {
                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("Cannot reset transaction timeout, tx in execution");
                }
                // Cannot reset transaction timeout, tx (ut.begin)in execution, ignore.
                return;
            }
        }
        
        if (timeout > 0) {
            threadTimeout.set(new Integer(timeout));
            // transactionTimeout = timeout;
        } else {
            threadTimeout.set(new Integer(0));
            // transactionTimeout = defaultTimeout;
        }
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Resetting transaction timeout= " + ((Integer) threadTimeout.get()).intValue());
        }
    }
    
    /**
     * Gets the transaction timeout value
     * 
     * @return transaction timeout value (in seconds)
     */

   public int getTransactionTimeout() {
    
      Integer mytimeobj = (Integer) threadTimeout.get();
      
      if (mytimeobj == null) {  // setTransactionTimeout not called, use defaultTimeout
          return defaultTimeout;
      } else {
          if (mytimeobj.intValue() == 0) {  // transactionTimeout not set, use defaultTimeout
              return defaultTimeout;   
          } else {
              return (mytimeobj.intValue());
          }
      }
      // return transactionTimeout;
   }
   
    /**
     * Modifies the value of the recovery value that is associated with the
     * transactions started by the current thread with the begin method.
     *
     * If an application has not called this method, the transaction
     * service uses the default value of 'false' for recovery.
     *
     * @param recovery value of the recovery (true or faluse). If the value is
     * false, recovery of transactions is disabled.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */

   public void setTransactionRecovery(boolean recovery) throws SystemException {

       if (TraceTm.recovery.isDebugEnabled()) {
           TraceTm.recovery.debug("recovery="+ recovery);
       }

       if (recovery) {
           transactionRecovery = recovery;
       } else {
           transactionRecovery = DEFAULT_RECOVERY;
       }
   }

    // ------------------------------------------------------------------
    // TransactionManager implementation
    // (only the methods that are not already in UserTransaction)
    // ------------------------------------------------------------------

    /**
     * Gets the transaction object that represents the transaction context of
     * the calling thread.
     *
     * @return the Transaction object representing the transaction 
     * associated with the calling thread. If the calling thread is
     * not associated with a transaction, a null object reference
     * is returned.
     *
     * @exception SystemException Thrown if the transaction manager 
     * encounters an unexpected error condition
     */

    public Transaction getTransaction() throws SystemException {
        Transaction ret = (Transaction) threadTx.get();

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.get= " + threadTx.toString());
            TraceTm.jta.debug("Transaction ret= " + ret);
        }

        return ret;
    }

    /**
     * Resumes the transaction context association of the calling thread with
     * the transaction represented by the supplied Transaction object. When this
     * method returns, the calling thread is associated with the transaction
     * context specified. 
     * <p><em>Warning</em>: No XA start is done here. We suppose it is already
     * done after a <code>getConnection()</code>. 
     * </p>
     * The supposed programming model is: <ol>
     * 	<li><code>getConnection()</code></li>
     * 	<li>SQL code</li>
     * 	<li><code>connection.close()</code</li>
     * 	</ol>
     *
     * @param tobj The <code>Transaction</code> object that represents the
     * transaction to be resumed.
     *
     * @exception InvalidTransactionException Thrown if the parameter
     * transaction object contains an invalid transaction
     *
     * @exception IllegalStateException Thrown if the thread is already
     * associated with another transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */

    public void resume(Transaction tobj) throws InvalidTransactionException, IllegalStateException, SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("resume transaction");
        }

        // invalid Transaction
        if (tobj == null) {
            TraceTm.jotm.error("resume: null arg.");
            throw new InvalidTransactionException("resume(null) is not valid");
        }

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("tx="+ tobj);
        }

        // Checks that the thread is not already associated to ANOTHER transaction
        Transaction mytx = (Transaction) threadTx.get();

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.get= " + threadTx.toString());
        }

        if (mytx != null) {
            if (mytx.equals(tobj)) {
                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("nothing to do");
                }
                return;
            }
            TraceTm.jotm.error("resume: already associated with another transaction.");
            throw new IllegalStateException("the thread is already associated with another transaction.");
        }

        // test for type before cast
        if (!(tobj instanceof TransactionImpl)) {
            TraceTm.jotm.error("resume: non TransactionImpl arg.");
            throw new InvalidTransactionException("resume(" + tobj.getClass().getName() + ") is not valid");
        }

        // Associates this Tx with the current thread
        threadTx.set(tobj);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.set= " + threadTx.toString());
        }

        if (!appServer) {
            // attach suspended resources
            try {
                ((TransactionImpl) tobj).doAttach(XAResource.TMRESUME);
            } catch (RollbackException e) {
                // never.
                TraceTm.jotm.error("RollbackException occured in resume()");
                throw new SystemException("RollbackException in occured in resume() " + e.getMessage());
            }
        }
    }

    /** 
      * Suspends the transaction currently associated with the calling thread
      * and return a <code>Transaction</code> object that represents the
      * transaction context being suspended. 
      * If the calling thread is not
      * associated with a transaction, the method returns
      * <code>null</code>. When this method returns, the calling thread is
      * associated with no transaction.
      <p><em>Warning</em>: No XA start is done here. We suppose it is already
      done after a <code>getConnection()</code>.
     * </p>
     * The supposed programming model is: <ol>
     * 	<li><code>getConnection()</code></li>
     * 	<li>SQL code</li>
     * 	<li><code>connection.close()</code</li>
     * 	</ol>
     * 
     * @return Transaction object representing the suspended transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */

    public Transaction suspend() throws SystemException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("suspend transaction");
        }

        /// TMSUSPEND should be defined in Transaction.
        TransactionImpl tx = (TransactionImpl) threadTx.get();

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.get= " + threadTx.toString());
        }

        if (tx != null) {

            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("tx="+ tx);
            }
            
            if (!appServer) {
                tx.doDetach(XAResource.TMSUSPEND);
            }
            
            threadTx.set(null);

            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("threadTx.set= null");
            }
        }

        return tx;
    }

    // ------------------------------------------------------------------
    // ResourceManagerEventListener implementation
    // Should be in UserTransaction only.
    // ------------------------------------------------------------------

    /** 
      * @see org.objectweb.transaction.jta.ResourceManagerEventListener#connectionOpened(org.objectweb.transaction.jta.ResourceManagerEvent)
      */

    public void connectionOpened(ResourceManagerEvent event) {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.connectionOpened " + this);
        }

        List list = null;
        Stack curStack = (Stack) eventListStack.get();
        // get thread local stack

        if (curStack == null) {
            // no stack yet : create one
            eventListStack.set(curStack = new Stack());
        } else { // pop list from current stack
            try {
                list = (List) curStack.pop();
            } catch (EmptyStackException e) {
                // ignore: might happen if thread used by session bean
            }
        }

        // no list yet: create one
        if (list == null)
            list = new Vector(1);

        // add event to list
        list.add(event);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("list.add(event) = " + event);
        }

        // push list into stack
        curStack.push(list);
    }

    /**
     * @see org.objectweb.transaction.jta.ResourceManagerEventListener#connectionClosed(org.objectweb.transaction.jta.ResourceManagerEvent)
     */

    public void connectionClosed(ResourceManagerEvent event) {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.connectionClosed, remove");
        }

        // remove event from list
        removeFromCurrentStack(event);
    }

    /**
     * @see org.objectweb.transaction.jta.ResourceManagerEventListener#connectionErrorOccured(org.objectweb.transaction.jta.ResourceManagerEvent)
     */

    public void connectionErrorOccured(ResourceManagerEvent event) {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.connectionErrorOccured");
        }

        removeFromCurrentStack(event); // remove event from list
    }

    private void removeFromCurrentStack(ResourceManagerEvent event) {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.removeFromCurrentStack " + event);
            TraceTm.jta.debug("Current = " + this);
        }

        Stack curStack = (Stack) eventListStack.get();
        // get thread local stack

        if (curStack == null) {
            // if no transaction currently executing (no begin), no curStack
            // has been put on the eventListStack so just return
            return;
        }

        try {

            List list = (List) curStack.peek();

            if (list != null) {
                // get list from stack top

                list.remove(event);
            }
        } catch (EmptyStackException e) {
            // no list of RM Events are known, just return
        }
    }

    //
    // objectweb transaction manager implementation
    //

    // thread local variable that stores all RM Events

    private transient ThreadLocal eventListStack = new ThreadLocal();

    /**
     * @see org.objectweb.transaction.jta.TransactionManager#pushThreadLocalRMEventList(java.util.List)
     */

    public void pushThreadLocalRMEventList(List eventList) {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.pushThreadLocalRMEventList");
        }

        Stack curStack = (Stack) eventListStack.get();

        if (curStack == null) {
            eventListStack.set(curStack = new Stack());
        }

        curStack.push(eventList);
    }

    /**
     * @see org.objectweb.transaction.jta.TransactionManager#popThreadLocalRMEventList()
     */

    public List popThreadLocalRMEventList() {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.popThreadLocalRMEventList");
        }

        Stack curStack = (Stack) eventListStack.get();
        return (List) curStack.pop();
    }

    // ------------------------------------------------------------------
    // Referenceable implementation
    // ------------------------------------------------------------------

    /**
     * Retrieves the <code>Reference</code> of this object. 
     * 
     * @return  The non-null <code>Reference</code> of this object.
     *  
     * @exception  NamingException  If a naming exception was encountered while retrieving the reference.  
     */

    public Reference getReference() throws NamingException {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.getReference()");
        }

        // create the reference
        Reference ref = new Reference(this.getClass().getName(), "org.objectweb.jotm.UserTransactionFactory", null);
        Integer i = (Integer) threadTimeout.get();

        if (i == null) {  // use defaultTimeout
            i = new Integer(defaultTimeout);   
        } else {
             if (i.intValue() == 0) {  // use defaultTimeout
                 i = new Integer(defaultTimeout);  
             }
        }
          
        // Integer i = new Integer(transactionTimeout);
        ref.add(new StringRefAddr("jotm.timeout", i.toString()));

        return ref;
    }

    // ------------------------------------------------------------------
    // Other public methods
    // ------------------------------------------------------------------

    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized in case of plain client.
     *
     * @return The <code>Current</code> object created 
     */

    public static Current getCurrent() {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.getCurrent unique= " + unique);
        }
        return unique;
    }
    
    // /**
    //  * Returns the unique instance of the class or <code>null</code> if not
    //  * initialized in case of plain client.
    //  *
    //  * @return The <code>Current</code> object created 
    //  */
    
    // public static Current getCurrent(TransactionFactory mytmfact) {
    //     if (unique == null) {
    //         unique = new Current(mytmfact);
    //     }
        
    //     if (TraceTm.jta.isDebugEnabled()) {
    //         TraceTm.jta.debug("Current.getCurrent unique= " + unique);
    //     }
    //     return unique;
    // }
        
    /**
     * Returns the TMFactory (in JTM)
     * 
     * @return TransactionFactory
     */

    public static TransactionFactory getJTM() {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.getJTM tm= " + tm);
        }
        if (tm == null) {
            TraceTm.jotm.error("Current: TMFactory is null!");
        }

        return tm;
    }
    
    /**
     * Returns the Transaction Recovery object
     * 
     * @return TransactionRecovery
     */

    public static TransactionRecovery getTransactionRecovery() {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Current.getTransactionRecovery tr= " + tr);
        }
        if (tr == null) {
            TraceTm.jotm.error("Current: Transaction Recovery is null!");
        }

        return tr;
    }
    
    /**
     * Sets the default timeout value
     * 
     * @param timeout timeout value (in seconds)
     */

    public void setDefaultTimeout(int timeout) {

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("timeout= " + timeout);
        }

        if (timeout != 0) {
            defaultTimeout = timeout;
            // transactionTimeout = defaultTimeout;  // Set transaction timeout = default
        }
        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("default timeout= " + defaultTimeout);
        }

    }

    /**
      * Gets the default timeout value
      * 
      * @return default timeout value (in seconds)
      */

    public int getDefaultTimeout() {

        return defaultTimeout;
    }
    
    /**
     * Sets the default recovery value
     * 
     * @param recovery recovery value (true or false)
     */

    public static void setDefaultRecovery(boolean recovery) {

        TraceTm.recovery.info("Jotm Recovery= " + recovery);

        transactionRecovery = recovery;
    }

    /**
      * Gets the default recovery value
      * 
      * @return default recovery value (true or false)
      */

    public static boolean getDefaultRecovery() {

        return transactionRecovery;
    }
    
    /**
     * Sets indicator if running with Application Server
     * 
     * @param appserver application server enabled value (true or false)
     */

    public static void setAppServer(boolean appserver) {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("Application Server enabled= " + appserver);
        }

        appServer = appserver;
    }

    /**
      * Gets the Application Server enabled value
      * 
      * @return application server enabled value (true or false)
      */

    public static boolean getAppServer() {

        return appServer;
    }
    /**
      * Associate to the current thread a transaction represented by its
      * transaction context.
      * This is used internally by the implicit propagation of the
      * transactional context: 
      * <ul>
      * <li>in the skeleton, before calling the request (<code>isReply =
      * 	false</code>)</li>
      *  <li>in    the stub, after receiving the reply (<code>isReply =
      * 	true</code>)</li>
      * </ul>
      * 
      * @param pctx TransactionContext
      * @param isReply <code>true</code> before calling a request,
      * <code>false</code> after receiving a reply
      */

    public void setPropagationContext(TransactionContext pctx, boolean isReply) {

        if (TraceTm.jotm.isDebugEnabled()) {
            TraceTm.jotm.debug("pctx=" + pctx + ", isReply=" + isReply);
        }

        if (pctx == null) { // never isReply when pctx=null

            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("detach tx");
            }

            TransactionImpl tx = (TransactionImpl) threadTx.get();

            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("threadTx.get= " + threadTx.toString());
            }

            if (tx != null) {
                // Free TransactionImpl, if not used only!!!
                // This is to avoid memory leaks.

                if (tx.isRemovable()) {
                    forgetTx(tx.getXid());
                }

                // Detaches thread from any transaction
                threadTx.set(null);

                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("threadTx.set= null");
                }
            }

            return;
        }

        // Get the Transaction matching this Xid if it exists
        Xid xid = pctx.getXid();
        TransactionImpl tx = getTxXid(xid);

        if (tx == null) {

            if (!isReply) { // create a Transaction object

                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("new Tx");
                }

                tx = new TransactionImpl(pctx);
                putTxXid(xid, tx);

                // sets the time stamp for the transaction
                Date myDate = new Date();
                tx.setTxDate(myDate);
            } else {

                // This may happen with Jeremie and distributed transactions with
                // no dammage.

                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("unknown tx:" + xid);
                }
            }

        } else {

            if (isReply) {
                // This thread is already associated to a transaction.
                // A proper PropagationContext exists yet, but it may have been changed
                // in case of distributed transaction.

                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("updating Xid=" + xid);
                }

                tx.updatePropagationContext(pctx);
            } else {

                if (TraceTm.jta.isDebugEnabled()) {
                    TraceTm.jta.debug("transaction already known:" + xid);
                }
            }
        }

        // Associates this Tx with the current thread (= resume)
        // In case of Jeremy, if isReply, we are not in the correct thread, so
        // never touch threadTx in that case!!!

       if (!isReply) {
            threadTx.set(tx);

            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("threadTx.set= " + threadTx.toString());
            }
       }
    }

    /**
      * Get the transaction context associated with the current thread or null
      * if the thread is not involved in a transaction.
      */

    public TransactionContext getPropagationContext(boolean hold) {

        if (TraceTm.jotm.isDebugEnabled()) {
            TraceTm.jotm.debug("hold=" + hold);
        }

        try {
            TransactionImpl tx = (TransactionImpl) getTransaction();

            if (tx != null) { // will return null if Tx not valid.
                return tx.getPropagationContext(hold);
            }
        } catch (SystemException e) {
            TraceTm.jotm.error("getPropagationContext system exception:", e);
        }

        return null;
    }

    /**
      * Forget all about this transaction.
      * References to <code>TransactionImpl</code> must be destroyed to allow
      * the garbage collector to free memory allocated to this transaction.
      * 
      * @param xid <code>Xid</code> of the transaction
      */

    public void forgetTx(Xid xid) {

        // Only clear threadTx if this thread is working on its Xid

        TransactionImpl txCur = (TransactionImpl) txXids.get (xid);

        if (txCur != null && txCur.equals((TransactionImpl) threadTx.get())) {
            threadTx.set(null);
            
            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("threadTx.set = null");
            }
        }
        removeTxXid(xid);
    }

    /**
      * Get the transaction referenced by Xid.
      * 
      * @param xid <code>Xid</code> of the transaction
      */

    public TransactionImpl getTxByXid(Xid xid) {

        TransactionImpl tx = (TransactionImpl) txXids.get (xid);
        return tx;
    }

    /**
      * Get the Xid's of all prepared transactions.
      * 
      * @return array of all Xids in the prepared state
      */

    public javax.transaction.xa.Xid[] getPreparedHeuristicXid() {

        Vector      xidlist;
        int         xidcount;
        javax.transaction.xa.Xid         xid = null;

        xidcount = txXids.size();    /* Xid count */

        if (xidcount == 0) {
            return null;
        }

        xidlist = new Vector();

        Set txXidsSet = txXids.keySet();

        synchronized(txXids) {

            Iterator txXidsIterator = txXidsSet.iterator();

            while (txXidsIterator.hasNext()){
                TransactionImpl tx = (TransactionImpl) txXids.get(txXidsIterator.next());

                try {

                    if (tx.getStatus() == Status.STATUS_PREPARED) {
                        xid = (javax.transaction.xa.Xid) tx.getXid();
                        xidlist.add(xid);
                    }
                } catch (SystemException e) {
                    TraceTm.jotm.error("getPreparedHeuristicsXid system exception:", e);
                }
            }
        }

        Vector mytxrecovered = JotmRecovery.getTxRecovered();
        
        return (javax.transaction.xa.Xid[]) xidlist.toArray();
    }

    /**
      * Get all Xid's associated with this transaction.
      * 
      * @return array of all Xids
      */

    public javax.transaction.xa.Xid[] getAllXid() {

        Vector      xidlist;
        int         xidcount;
        javax.transaction.xa.Xid         xid = null;

        xidcount = txXids.size();   /* Xid count */

        if (xidcount == 0) {
            return null;
        }

        xidlist = new Vector();

        Set txXidsSet = txXids.keySet();

        synchronized(txXids) {

            Iterator txXidsIterator = txXidsSet.iterator();

            while (txXidsIterator.hasNext()){
                TransactionImpl mytx = (TransactionImpl) txXids.get(txXidsIterator.next());
                xid = (javax.transaction.xa.Xid) mytx.getXid();
                xidlist.add(xid);
            }
        }

        return (javax.transaction.xa.Xid[]) xidlist.toArray();
    }

    /**
      * Get all executing transactions.
      * 
      * @return array of all Transactions in execution
      */

    public String [] getAllTx() {

        Vector      txList;
        List        txResourceList;
        int         txCount;
        int         txResourceCount;
        String      txStatusName;

        txCount = txXids.size();   /* transaction count */

        if (txCount == 0) {
            return null;
        }

        txList = new Vector();

        Set txXidsSet = txXids.keySet();

        synchronized(txXids) {

            Iterator txXidsIterator = txXidsSet.iterator();

            while (txXidsIterator.hasNext()){
                TransactionImpl mytx = (TransactionImpl) txXids.get(txXidsIterator.next());

                try {
                     txStatusName = StatusHelper.getStatusName (mytx.getStatus());
                } catch (SystemException e) {
                     txStatusName = "No State Defined";
                }

                txResourceList = mytx.getEnlistedXAResource();
                txResourceCount = txResourceList.size();

                if (txResourceCount == 0) {
                    txList.add(mytx.getTxDate().toString() + "????" +
                               mytx.toString() + "????" +
                               "NO Resource Defined" + "????" +
                               txStatusName);
                } else {

                    for (int i=0; i < txResourceCount; i++) {
                         txList.add(mytx.getTxDate().toString() + "????" +
                                    mytx.toString() + "????" +
                                    txResourceList.get(i).toString() + "????" +
                                    txStatusName);
                    }
                }
            }
        }

        String [] myTxString = new String [txCount] ;

        for (int i=0; i < txCount; i++) {
            myTxString[i] = txList.get(i).toString();
        }

        return myTxString;
    }

    /**
     * Get all Transactions that may require recovery.
     * 
     * @return array of all Transactions that may require recovery
     */

   public String [] getAllRcTx() {

       Vector vTxRecovered = null;
       Vector      txList;
       int         txCount;

       JotmRecovery myjr = null;

       if (tr == null) {
           if (TraceTm.recovery.isDebugEnabled()) {
               TraceTm.recovery.debug("tr= null");
           }
           return null;
       }

       myjr = tr.getJotmRecovery();
       
       if (myjr == null) {
           return null;
       }

       vTxRecovered = JotmRecovery.getTxRecovered();
       
       txCount = vTxRecovered.size();
       
       if (TraceTm.recovery.isDebugEnabled()) {
          TraceTm.recovery.debug("txcount= " + txCount);
       }
       
       if (txCount == 0) {
           return null;
       }
       
       txList = new Vector();
       TxRecovered mytxRecovered = null;

       for (int i=0; i<txCount; i++){
           mytxRecovered = (TxRecovered) vTxRecovered.elementAt(i);
           
           Xid temptxxid = new XidImpl(mytxRecovered.gettxxid());
           
           // We must send the complete transaction Xid string along
           // with the truncated. The complete will be used for further
           // searching while the truncated will be used in the display
           // of the JSP.
           
           txList.add(new String(mytxRecovered.gettxxid()) + "????" +
                temptxxid.toString() + "????" +
                mytxRecovered.gettxdatetime() + "????" +
                mytxRecovered.getxidcount());
       }

       String [] myTxString = new String [txCount] ;

       for (int i=0; i < txCount; i++) {
           myTxString[i] = txList.get(i).toString();
       }

       return myTxString;
   }
   
   /**
    * Get all XAResources that may require recovery.
    * 
    * @return array of all XAResources that may require recovery
    */

  public String [] getAllXaTx(String stx) {

      Vector      xaList;
      Vector      rmList;
      int         txCount;
      int         rmiCount;
      int         myxacount = 0;
      boolean     mytxfound = false;
      boolean     myrmifound = false;
      boolean     myrmregistered = false;
      boolean     myxaresfound = false;
      String      mytxstatusname;

      JotmRecovery myjr = null;

      if (tr == null) {
          return null;
      }

      myjr = tr.getJotmRecovery();
      
      if (myjr == null) {
          if (TraceTm.recovery.isDebugEnabled()) {
              TraceTm.recovery.debug("myjr= null");
          }
          return null;
      }
      
      xaList = new Vector();
      TxxidRecovered mytxxidRecovered = null;
      TxRecovered mytxRecovered = null;
      Vector vTxRecovered = null;
      Vector vRecoverRmInfo = null;
      RmRegistration myRmRegistration = null;
      String myxares = null;
      
      vTxRecovered = JotmRecovery.getTxRecovered();

      txCount = vTxRecovered.size();
      
      if (TraceTm.recovery.isDebugEnabled()) {
          TraceTm.recovery.debug("txcount= " + txCount);
      }
      
      for (int i=0; i<txCount; i++){
          mytxRecovered = (TxRecovered) vTxRecovered.elementAt(i);
          
          if (TraceTm.recovery.isDebugEnabled()) {
              TraceTm.recovery.debug("gettxxid= " + new String(mytxRecovered.gettxxid()));
              TraceTm.recovery.debug("stx=" + stx);
          }
          
          if (new String(mytxRecovered.gettxxid()).equals(stx)) {
              mytxfound = true;
              break;
          }
      }
      
      rmList = new Vector();
      RecoverRmInfo rmInfo = null;
      vRecoverRmInfo = JotmRecovery.getRecoverRmInfo();
      String myrm = null;
      Vector vRmRegistration = tr.getRmRegistration();
      
      if (mytxfound) {
          myxacount = mytxRecovered.getxidcount();
          
          if (TraceTm.recovery.isDebugEnabled()) {
              TraceTm.recovery.debug("myxacount= " + myxacount);
          }
          
          for (int i=0; i<myxacount; i++) {
              myrmregistered = false;
              myrmifound = false;
              myxaresfound = false;
              mytxxidRecovered = mytxRecovered.getRecoverTxXidInfo(i);

              if (mytxxidRecovered == null) {
                  xaList.add("NotFound" + "????" +
                             "NotFound" + "????" +
                             "NotFound" + "????" +
                             "NotFound" + "????" +
                             "NotFound");
              } else {
                  rmiCount = vRecoverRmInfo.size();
                  
                  if (TraceTm.recovery.isDebugEnabled()) {
                      TraceTm.recovery.debug("rmiCount= " + rmiCount);
                  }
                
                  for (int j=0; j<rmiCount; j++){
                      rmInfo = (RecoverRmInfo) vRecoverRmInfo.elementAt(j);
                      
                      if (TraceTm.recovery.isDebugEnabled()) {
                          TraceTm.recovery.debug("getRecoverXaResName()= " + rmInfo.getRecoverXaResName());
                          TraceTm.recovery.debug("getRecoverxaresname()=" + mytxxidRecovered.getRecoverxaresname());
                          TraceTm.recovery.debug("getRecoverXaRes()= " + new String(rmInfo.getRecoverXaRes()));
                          TraceTm.recovery.debug("getRecoverxares()=" + new String(mytxxidRecovered.getRecoverxares()));
                      }
                     
                      // if (rmInfo.getRecoverXaResName().equals(mytxxidRecovered.getRecoverxaresname())) {
                      if ((new String(rmInfo.getRecoverXaRes())).equals(new String(mytxxidRecovered.getRecoverxares()))) {
                          myrm = rmInfo.getRecoverRm();
                          myrmifound = true;
                          myxares = new String(mytxxidRecovered.getRecoverxares());
                          
                          if (TraceTm.recovery.isDebugEnabled()) {
                              TraceTm.recovery.debug("myrm= " + myrm);
                              TraceTm.recovery.debug("myxares= " + myxares);
                          }
                          break;
                      }
                  }
              
                  if (!myrmifound) {
                      myrm = "NotFound";
                  }
                    
                  if (vRmRegistration == null) {
                      if (TraceTm.recovery.isDebugEnabled()) {
                          TraceTm.recovery.debug("vRmRegistration is null");
                      }
                      myxares = "NotRegistered";
                  } else {
                      int rmregcount = vRmRegistration.size();
                      
                      if (TraceTm.recovery.isDebugEnabled()) {
                          TraceTm.recovery.debug("rmregcount= " + rmregcount);
                      }
                      
                      for (int j=0; j<rmregcount; j++){
                          myRmRegistration = (RmRegistration) vRmRegistration.elementAt(j);
                          
                          if (TraceTm.recovery.isDebugEnabled()) {
                              TraceTm.recovery.debug("myrm= " + myrm);
                              TraceTm.recovery.debug("rmGetName= " + myRmRegistration.rmGetName());
                          }
                          
                          if (myrm.equals(myRmRegistration.rmGetName())) {
                              myrmregistered = true;
                              
                              if (myRmRegistration.rmGetXaRes() == null) {
                                  myxares = "IsNull";   
                              } else {
                                  myxares = myRmRegistration.rmGetXaRes().toString();
                                  myxaresfound = true;
                                  
                                  if (TraceTm.recovery.isDebugEnabled()) {
                                      TraceTm.recovery.debug("myxares= " + myxares);
                                  }
                              }
                              break;
                          }
                      }
                        
                      if (!myrmregistered) {
                          myxares = "NotRegistered";   
                      }
                      
                      if (TraceTm.recovery.isDebugEnabled()) {
                          TraceTm.recovery.debug("myxares= " + myxares);
                      }
                  }
                  
                  Xid tempxid = new XidImpl(mytxxidRecovered.getRecoverxid());
                  
                  // We must send the complete Xid string along with the truncated.
                  // The complete will be used for further searching while the
                  // truncated will be used in the display of the JSP.
                  xaList.add(myrm + "????" +
                        myxares + "????" +
                        new String(mytxxidRecovered.getRecoverxid()) + "????" +
                        new String(tempxid.toString()) + "????" +
                        StatusHelper.getStatusName (mytxxidRecovered.getRecoverstatus()));
              }
          }
      }

      String [] myTxString = new String [myxacount] ;

      for (int i=0; i < myxacount; i++) {
          myTxString[i] = xaList.get(i).toString();
      }

      return myTxString;
  }
  
  /**
   * @return Returns all XAResources that require administrator recovery action.
   */
  public int actionXAResource (String xaAction, String xatx) {
      int commiterror;
      
      Vector         vRmRegistration = null;
      Vector         vRecoverRmInfo = null;
      int            rmregcount;
      boolean        mytxfound = false;
      boolean        myrmifound = false;
      String         mytxstatusname;
      JotmRecovery   myjr = null;
      RmRegistration myRmRegistration = null;
      XAResource     xaresource = null;

      if (tr == null) {
          return 0;
      }

      myjr = tr.getJotmRecovery();
      
      if (myjr == null) {
          return 0;
      }

      vRmRegistration = tr.getRmRegistration();
      
      if (vRmRegistration == null) {
           return 0;
      }
      
      // Given the passed Resource Manager, find it in the RMRegistered vector.
      // When the Resource Manager is found, use the XAResource associated with
      // the RMRegistered Resource Manager to perform a XAResource.recover call
      // to return all XIDs that require recovery.
      // With the list of XIDs to be recovered, verify that the Xid passes in
      // the list. If the Xid passed is in the recover list, then we can attempt
      // the commit of the Xid (i.e., XAResource.Commit(Xid)).
      // When the commit is completed, we can remove the TxXidRecovered array
      // entry associated with the Xid and decrement the count in its respective
      // TxRecovered vector entry. If the count in the TxRecovered vector entry
      // is zero, we can then remove the TxRecovered entry from the vector.
      
      String mys = xatx;
      int myix1 = mys.indexOf('\n');
      String sResmgr = mys.substring(0, myix1);
      int myix2 = mys.indexOf('\n', myix1 + 1);
      String sResource = mys.substring(myix1 + 1, myix2);
      int myix3 = mys.indexOf('\n', myix2 + 1);
      String sFullXid = mys.substring(myix2 + 1, myix3);
      String sXidstate = mys.substring(myix3 + 1);

      rmregcount = vRmRegistration.size();
      
      for (int i=0; i<rmregcount; i++){
        myRmRegistration = (RmRegistration) vRmRegistration.elementAt(i);

        if (sResmgr.equals(myRmRegistration.rmGetName())) {
            xaresource = myRmRegistration.rmGetXaRes();
            break;
        }
      }
      
      // int rcflag = 0;
      
      // javax.transaction.xa.Xid javaxid [] = null; // new javax.transaction.xa.Xid[100];
      
      List recoveredXidList = new LinkedList();  // new code
      
      try {
          // recover returns javax.transaction.xa.Xid objects
          // javaxid = xaresource.recover(rcflag);
          boolean first = true;
          while (true) {
              javax.transaction.xa.Xid[] javaxids;
              javaxids = xaresource.recover (first ? XAResource.TMSTARTRSCAN : XAResource.TMNOFLAGS);
              if (javaxids == null || javaxids.length == 0)
                  break;
           
              first = false;
              recoveredXidList.addAll(Arrays.asList(javaxids));
          }
      } catch (XAException e) {
          if (TraceTm.jta.isDebugEnabled()) {
              TraceTm.recovery.debug("xaResource.recover call failed during recovery " + e.getMessage());
          }
      }
      
      // if (javaxid == null) {

      //     if (TraceTm.recovery.isDebugEnabled()) {
      //         TraceTm.recovery.debug("No XIDs to recover for Xares javaxid is null");
      //     }
          
      //     cleanuptxrecovery(sFullXid);
      //     return 0; 
      // }
      
      // if (javaxid.length == 0) {
      if (recoveredXidList.size() == 0) {

          if (TraceTm.recovery.isDebugEnabled()) {
              TraceTm.recovery.debug("No XIDs to recover for Xares= "+ xaresource);
          }
          
          cleanuptxrecovery(sFullXid);
          return 0;
      }

      // Xid rcxid [] = new Xid[javaxid.length];
      Iterator xidIter = recoveredXidList.iterator();
      
      // for (int xx=0; xx<javaxid.length; xx++) {
      while (xidIter.hasNext()) {
          Xid recoveredXid = (Xid) xidIter.next();
      	
      // 	  if (javaxid[xx] == null) break;

      //     if (new String(javaxid[xx].toString()).equals(sFullXid)) {
          if (new String(recoveredXid.toString()).equals(sFullXid)) {
            
              // if (xaAction == "commit") {
              if (xaAction.equals("commit")) {
                  try {
                      // xaresource.commit(javaxid[xx], false);
                      xaresource.commit(recoveredXid, false);
                  } catch (XAException e) {
                      TraceTm.recovery.error("Unable to commit Xid during Admin Recovery " + e.getMessage());
                  }
              } else {
                  // if (xaAction == "rollback") {
                  if (xaAction.equals("rollback")) {
                      try {
                          // xaresource.rollback(javaxid[xx]);
                          xaresource.rollback(recoveredXid);
                      } catch (XAException e) {
                          TraceTm.recovery.error("Unable to rollback Xid during Admin Recovery " + e.getMessage());
                      }
                  } else {
                       // if (xaAction == "forget") {
                       if (xaAction.equals("forget")) {
                           try {
                               // xaresource.rollback(javaxid[xx]);
                               xaresource.rollback(recoveredXid);
                           } catch (XAException e) {
                               TraceTm.recovery.error("Unable to rollback Xid during Admin Recovery " + e.getMessage());
                           }
                       }
                  }
              }
          }
          break;
      }
      
      cleanuptxrecovery(sFullXid);
      return 0;
  }
  
  private void cleanuptxrecovery(String pFullXid) {
      
      // We can now remove the TxXidRecovered array entry associated with the
      // Xid and decrement the count in its respective TxRecovered vector entry.
      // If the count in the TxRecovered vector entry is zero, we can then remove
      // the TxRecovered entry from the vector.

      Vector vTxRecovered = null;
      TxRecovered mytxRecovered = null;
      TxxidRecovered mytxxidRecovered = null;
      int         txCount;
      int         myxacount = 0;
      boolean mytxxidrecoveredfound = false;
    
      byte [] jotmDone = new byte [11];
      byte [] [] jotmDoneRecord = new byte [1] [11];

      jotmDone = "RR3JOTMDONE".getBytes();
      
      vTxRecovered = JotmRecovery.getTxRecovered();
      txCount = vTxRecovered.size();
      
      for (int i=0; i<txCount; i++){
          mytxRecovered = (TxRecovered) vTxRecovered.elementAt(i);
          myxacount = mytxRecovered.getxidcount();
          
          for (int j=0; j<myxacount; j++) {
              mytxxidRecovered = mytxRecovered.getRecoverTxXidInfo(j);
              
              if (mytxxidRecovered != null) {
                
                  if (pFullXid.equals(new String(mytxxidRecovered.getRecoverxid()))) {
                      // mytxxidRecovered.setRecoverstatus(Status.STATUS_NO_TRANSACTION);
                      mytxxidRecovered.setRecoverstatus(Status.STATUS_COMMITTED);
                      mytxxidrecoveredfound = true;
                      break;
                  }
              }
          }
          
          boolean allcompleted = true;
          
          for (int j=0; j<myxacount; j++) {
              mytxxidRecovered = mytxRecovered.getRecoverTxXidInfo(j);
            
              // if (mytxxidRecovered.getRecoverstatus() != Status.STATUS_NO_TRANSACTION) {  // something other than completed
              if (mytxxidRecovered.getRecoverstatus() != Status.STATUS_COMMITTED) {  // something other than completed
                  allcompleted = false;
                  break;
              }
          }
          
          if (allcompleted) {
              XACommittingTx xaCommitTx = mytxRecovered.getXACommittingTx();
              jotmDoneRecord [0] = jotmDone;
            
              if (Current.getDefaultRecovery()) {
                  try {
                      if (TraceTm.recovery.isDebugEnabled()) {
                          TraceTm.recovery.debug("Done howl log, after admin action");
                      }
                    
                      TransactionRecoveryImpl.getTransactionRecovery().howlDoneLog(jotmDoneRecord, xaCommitTx);
                  } catch (Exception f) {
                      String howlerror =
                          "Cannot howlDoneLog:"
                          + f
                          + "--"
                          + f.getMessage();
                      TraceTm.jotm.error("Got LogException from howlDoneLog: "+ howlerror);
                  }
              }
            
              vTxRecovered.remove(i);
              break;
          }
          
          if (mytxxidrecoveredfound) return;
      }
  }
  
    /**
     * Associate Thread to this transaction.
     * (used by iiop interceptor)
     */

    public void associateThreadTx(Xid xid) {
  
        TransactionImpl tx = getTxXid(xid);
        threadTx.set(tx);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.set= " + threadTx.toString());
        }
    }

    /**
     * Clear transaction from this thread if not known.
     * Useful when another thread completes the current thread's transaction
     */
    public void clearThreadTx() {
  
        TransactionImpl tx = (TransactionImpl) threadTx.get();

        if (tx != null) {
            threadTx.set(null);
            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("threadTx.set=null");
            }
        }
    }

    // ------------------------------------------------------------------
    // private methods
    // ------------------------------------------------------------------

    /*
     * put the Tx/Xid mapping into the hashtable.
     */

    private void  putTxXid( Xid xid, TransactionImpl tx ) {

        if ( TraceTm.jta.isDebugEnabled() ) {
            TraceTm.jta.debug("Associate tx to xid (xid=" + xid + ")");
        }

        txXids.put(xid, tx);
    }

    /*
     * given the Xid, get the corresponding Tx from the hashtable.
     */

    private TransactionImpl  getTxXid( Xid xid ) {

        if ( TraceTm.jta.isDebugEnabled() ) {
            TraceTm.jta.debug("get tx from xid (xid="+ xid +")");
        }

        TransactionImpl tx = (TransactionImpl) txXids.get(xid);
        return tx;
    }

    /*
     * removeTxXid method.
     */

    private void  removeTxXid( Xid xid ) {

        if ( TraceTm.jta.isDebugEnabled() ) {
            TraceTm.jta.debug("remove tx from xid (xid="+ xid +")");
        }

        txXids.remove (xid);
    }

    /**
     * remove the Transaction from the ThreadLocal
     */

    void forget() throws Exception {
        threadTx.set(null);

        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("threadTx.set= null");
        }
    }

    /* Management methods */

    /**
      * Returns the current number of transactions.
      * 
      * @return current number of transaction
      */

    public int getTotalCurrentTransactions() {
        return txXids.size();
    }

    /**
     * Increments number of begun transactions by one.
     */

    synchronized void incrementBeginCounter() {
        nb_bg_tx++;
    }

    /**
     * Returns the total number of begun transactions.
     * 
     * @return total number of begun transactions
     */

    public int getTotalBegunTransactions() {
        return nb_bg_tx;
    }

    /**
     * Increments the number of rolled back transaction by one.
     */

    synchronized void incrementRollbackCounter() {
        nb_rb_tx++;
    }

    /**
     * Returns the total number of rolled back transactions.
     * 
     * @return total number of rolled back transactions
     */

    public int getTotalRolledbackTransactions() {
        return nb_rb_tx;
    }

    /**
     * Increments the number of of committed transactions by one.
     */

    synchronized void incrementCommitCounter() {
        nb_cm_tx++;
    }

    /**
     * Returns the total number of committed transactions.
     *
     * @return total number of commited transactions
     */

    public int getTotalCommittedTransactions() {
        return nb_cm_tx;
    }

    /**
     * Resets total number of transactions.
     */

    public synchronized void resetAllTxTotalCounters() {
        nb_bg_tx = 0;
        nb_cm_tx = 0;
        nb_rb_tx = 0;
        nb_to = 0;
    }

    /**
     * Increments number of rolled back transactions due to timeout by one.
     */

    synchronized void incrementExpiredCounter() {
        nb_to++;
    }

    /**
     * Returns the total number of  rolled back transactions due to timeout.
     * 
     * @return number of rolled back transactions due to timeout
     */

    public int getTotalExpiredTransactions() {
        return nb_to;
    }

    /**
     * Returns all counters.
     * 
     * @return an array of all counters (current tx, begun tx, committed tx,
     * rolled back tx, timeouted tx)
     */

    public synchronized Integer[] getTransactionCounters() {
        Integer[] result = new Integer[5];
        result[0] = new Integer(txXids.size());
        result[1] = new Integer(nb_bg_tx);
        result[2] = new Integer(nb_cm_tx);
        result[3] = new Integer(nb_rb_tx);
        result[4] = new Integer(nb_to);
        return result;
    }

}
